@@ -1,11 +1,36 @@
+from __future__ import absolute_import, division, print_function
+
+__copyright__ = """
+Copyright (C) 2009-2017 Andreas Kloeckner
+Copyright (C) 2014-2017 Aaron Meurer
+"""
+
+__license__ = """
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in
+all copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+THE SOFTWARE.
+"""
+
+
 import os
 import sys
 
-from pudb.py3compat import PY3
-if PY3:
-    from configparser import ConfigParser
-else:
-    from ConfigParser import ConfigParser
+from pudb.py3compat import ConfigParser
+from pudb.lowlevel import lookup_module, get_breakpoint_invalid_reason
 
 # minor LGPL violation: stolen from python-xdg
 
@@ -13,27 +38,31 @@ _home = os.environ.get('HOME', None)
 xdg_data_home = os.environ.get('XDG_DATA_HOME',
             os.path.join(_home, '.local', 'share') if _home else None)
 
-xdg_config_home = os.environ.get('XDG_CONFIG_HOME',
-            os.path.join(_home, '.config') if _home else None)
 
-xdg_config_dirs = [xdg_config_home] if xdg_config_home else [] + \
-    os.environ.get('XDG_CONFIG_DIRS', '/etc/xdg').split(':')
+XDG_CONFIG_HOME = os.environ.get('XDG_CONFIG_HOME',
+                                 os.path.join(_home, '.config') if _home else None)
+
+if XDG_CONFIG_HOME:
+    XDG_CONFIG_DIRS = [XDG_CONFIG_HOME]
+else:
+    XDG_CONFIG_DIRS = os.environ.get('XDG_CONFIG_DIRS', '/etc/xdg').split(':')
 
 
 def get_save_config_path(*resource):
-    if xdg_config_home is None:
+    if XDG_CONFIG_HOME is None:
         return None
     if not resource:
         resource = [XDG_CONF_RESOURCE]
     resource = os.path.join(*resource)
     assert not resource.startswith('/')
-    path = os.path.join(xdg_config_home, resource)
+    path = os.path.join(XDG_CONFIG_HOME, resource)
     if not os.path.isdir(path):
         os.makedirs(path, 448)  # 0o700
     return path
 
 # end LGPL violation
 
+
 CONF_SECTION = "pudb"
 XDG_CONF_RESOURCE = "pudb"
 CONF_FILE_NAME = "pudb.cfg"
@@ -51,11 +80,11 @@ def load_config():
     try:
         cparser.read([
             join(cdir, XDG_CONF_RESOURCE, CONF_FILE_NAME)
-            for cdir in xdg_config_dirs if isdir(cdir)])
+            for cdir in XDG_CONFIG_DIRS if isdir(cdir)])
 
         if cparser.has_section(CONF_SECTION):
             conf_dict.update(dict(cparser.items(CONF_SECTION)))
-    except:
+    except Exception:
         pass
 
     conf_dict.setdefault("shell", "internal")
@@ -74,25 +103,30 @@ def load_config():
 
     conf_dict.setdefault("custom_theme", "")
     conf_dict.setdefault("custom_stringifier", "")
+    conf_dict.setdefault("custom_shell", "")
 
     conf_dict.setdefault("wrap_variables", True)
+    conf_dict.setdefault("default_variables_access_level", "public")
 
     conf_dict.setdefault("display", "auto")
 
     conf_dict.setdefault("prompt_on_quit", True)
 
+    conf_dict.setdefault("hide_cmdline_win", False)
+
     def normalize_bool_inplace(name):
         try:
             if conf_dict[name].lower() in ["0", "false", "off"]:
                 conf_dict[name] = False
             else:
                 conf_dict[name] = True
-        except:
+        except Exception:
             pass
 
     normalize_bool_inplace("line_numbers")
     normalize_bool_inplace("wrap_variables")
     normalize_bool_inplace("prompt_on_quit")
+    normalize_bool_inplace("hide_cmdline_win")
 
     return conf_dict
 
@@ -113,7 +147,7 @@ def save_config(conf_dict):
         outf = open(join(save_path, CONF_FILE_NAME), "w")
         cparser.write(outf)
         outf.close()
-    except:
+    except Exception:
         pass
 
 
@@ -124,17 +158,18 @@ def edit_config(ui, conf_dict):
 
     def _update_theme():
         ui.setup_palette(ui.screen)
-
-        for sl in ui.source:
-            sl._invalidate()
+        ui.screen.clear()
 
     def _update_line_numbers():
         for sl in ui.source:
-                sl._invalidate()
+            sl._invalidate()
 
     def _update_prompt_on_quit():
         pass
 
+    def _update_hide_cmdline_win():
+        ui.update_cmdline_win()
+
     def _update_current_stack_frame():
         ui.update_stack()
 
@@ -143,6 +178,9 @@ def edit_config(ui, conf_dict):
         pudb.var_view.custom_stringifier_dict = {}
         ui.update_var_view()
 
+    def _update_default_variables_access_level():
+        ui.update_var_view()
+
     def _update_wrap_variables():
         ui.update_var_view()
 
@@ -170,6 +208,11 @@ def edit_config(ui, conf_dict):
             conf_dict.update(new_conf_dict)
             _update_prompt_on_quit()
 
+        elif option == "hide_cmdline_win":
+            new_conf_dict["hide_cmdline_win"] = not check_box.get_state()
+            conf_dict.update(new_conf_dict)
+            _update_hide_cmdline_win()
+
         elif option == "current_stack_frame":
             # only activate if the new state of the radio button is 'on'
             if new_state:
@@ -185,6 +228,13 @@ def edit_config(ui, conf_dict):
 
                 conf_dict.update(stringifier=newvalue)
                 _update_stringifier()
+
+        elif option == "default_variables_access_level":
+            # only activate if the new state of the radio button is 'on'
+            if new_state:
+                conf_dict.update(default_variables_access_level=newvalue)
+                _update_default_variables_access_level()
+
         elif option == "wrap_variables":
             new_conf_dict["wrap_variables"] = not check_box.get_state()
             conf_dict.update(new_conf_dict)
@@ -195,7 +245,7 @@ def edit_config(ui, conf_dict):
         "Configuration settings are saved in "
         "$HOME/.config/pudb or $XDG_CONFIG_HOME/pudb "
         "environment variable. If both variables are not set "
-        " configurations settings will not be saved.\n")
+        "configurations settings will not be saved.\n")
 
     cb_line_numbers = urwid.CheckBox("Show Line Numbers",
             bool(conf_dict["line_numbers"]), on_state_change=_update_config,
@@ -205,17 +255,37 @@ def edit_config(ui, conf_dict):
             bool(conf_dict["prompt_on_quit"]), on_state_change=_update_config,
                 user_data=("prompt_on_quit", None))
 
+    hide_cmdline_win = urwid.CheckBox("Hide command line (Ctrl-X) window "
+                                      "when not in use",
+            bool(conf_dict["hide_cmdline_win"]), on_state_change=_update_config,
+                user_data=("hide_cmdline_win", None))
+
     # {{{ shells
 
     shell_info = urwid.Text("This is the shell that will be "
             "used when you hit '!'.\n")
-    shells = ["internal", "classic", "ipython", "bpython", "ptpython"]
+    shells = ["internal", "classic", "ipython", "bpython", "ptpython", "ptipython"]
+    known_shell = conf_dict["shell"] in shells
+    shell_edit = urwid.Edit(edit_text=conf_dict["custom_shell"])
+    shell_edit_list_item = urwid.AttrMap(shell_edit, "value")
 
     shell_rb_group = []
     shell_rbs = [
             urwid.RadioButton(shell_rb_group, name,
                 conf_dict["shell"] == name)
-            for name in shells]
+            for name in shells]+[
+                urwid.RadioButton(shell_rb_group, "Custom:",
+                not known_shell, on_state_change=_update_config,
+                user_data=("shell", None)),
+                shell_edit_list_item,
+                urwid.Text("\nTo use a custom shell, see example-shell.py "
+                    "in the pudb distribution. Enter the full path to a "
+                    "file like it in the box above. '~' will be expanded "
+                    "to your home directory. The file should contain a "
+                    "function called pudb_shell(_globals, _locals) "
+                    "at the module level. See the PuDB documentation for "
+                    "more information."),
+            ]
 
     # }}}
 
@@ -297,6 +367,26 @@ def edit_config(ui, conf_dict):
 
     # }}}
 
+    # {{{ variables access level
+
+    default_variables_access_level_opts = ["public", "private", "all"]
+    default_variables_access_level_rb_group = []
+    default_variables_access_level_info = urwid.Text(
+            "Set the default attribute visibility "
+            "of variables in the variables list.\n"
+            "\nNote that you can change this option on "
+            "a per-variable basis by selecting the "
+            "variable and pressing '*'.")
+    default_variables_access_level_rbs = [
+            urwid.RadioButton(default_variables_access_level_rb_group, name,
+                conf_dict["default_variables_access_level"] == name,
+                on_state_change=_update_config,
+                user_data=("default_variables_access_level", name))
+            for name in default_variables_access_level_opts
+            ]
+
+    # }}}
+
     # {{{ wrap variables
 
     cb_wrap_variables = urwid.CheckBox("Wrap variables",
@@ -333,11 +423,10 @@ def edit_config(ui, conf_dict):
 
     lb_contents = (
             [heading]
-            + [urwid.AttrMap(urwid.Text("Line Numbers:\n"), "group head")]
+            + [urwid.AttrMap(urwid.Text("General:\n"), "group head")]
             + [cb_line_numbers]
-
-            + [urwid.AttrMap(urwid.Text("\nPrompt on quit:\n"), "group head")]
             + [cb_prompt_on_quit]
+            + [hide_cmdline_win]
 
             + [urwid.AttrMap(urwid.Text("\nShell:\n"), "group head")]
             + [shell_info]
@@ -354,6 +443,11 @@ def edit_config(ui, conf_dict):
             + [stringifier_info]
             + stringifier_rbs
 
+            + [urwid.AttrMap(urwid.Text("\nVariables Attribute Visibility:\n"),
+                "group head")]
+            + [default_variables_access_level_info]
+            + default_variables_access_level_rbs
+
             + [urwid.AttrMap(urwid.Text("\nWrap Variables:\n"), "group head")]
             + [cb_wrap_variables]
             + [wrap_variables_info]
@@ -385,9 +479,13 @@ def edit_config(ui, conf_dict):
             conf_dict.update(stringifier=newvalue, custom_stringifier=newvalue)
             _update_stringifier()
 
-        for shell, shell_rb in zip(shells, shell_rbs):
-            if shell_rb.get_state():
-                conf_dict["shell"] = shell
+        if shell_rb_group[-1].state:
+            newvalue = shell_edit.get_edit_text()
+            conf_dict.update(shell=newvalue, custom_shell=newvalue)
+        else:
+            for shell, shell_rb in zip(shells, shell_rbs):
+                if shell_rb.get_state():
+                    conf_dict["shell"] = shell
 
         for display, display_rb in zip(displays, display_rbs):
             if display_rb.get_state():
@@ -428,9 +526,7 @@ def parse_breakpoints(lines):
         if colon > 0:
             filename = arg[:colon].strip()
 
-            from pudb.lowlevel import lookup_module
             f = lookup_module(filename)
-
             if not f:
                 continue
             else:
@@ -444,7 +540,6 @@ def parse_breakpoints(lines):
         else:
             continue
 
-        from pudb.lowlevel import get_breakpoint_invalid_reason
         if get_breakpoint_invalid_reason(filename, lineno) is None:
             breakpoints.append((filename, lineno, False, cond, funcname))
 
@@ -461,35 +556,39 @@ def get_breakpoints_file_name():
 
 
 def load_breakpoints():
+    """
+    Loads and check saved breakpoints out from files
+    Returns: list of tuples
+    """
     from os.path import join, isdir
 
-    file_names = [
-            join(cdir, XDG_CONF_RESOURCE, name)
-            for cdir in xdg_config_dirs if isdir(cdir)
-            for name in [SAVED_BREAKPOINTS_FILE_NAME, BREAKPOINTS_FILE_NAME]
-            ]
+    file_names = []
+    for cdir in XDG_CONFIG_DIRS:
+        if isdir(cdir):
+            for name in [SAVED_BREAKPOINTS_FILE_NAME, BREAKPOINTS_FILE_NAME]:
+                file_names.append(join(cdir, XDG_CONF_RESOURCE, name))
 
     lines = []
     for fname in file_names:
         try:
-            rcFile = open(fname)
+            rc_file = open(fname, "rt")
         except IOError:
             pass
         else:
-            lines.extend([l.strip() for l in rcFile.readlines()])
-            rcFile.close()
+            lines.extend([l.strip() for l in rc_file.readlines()])
+            rc_file.close()
 
     return parse_breakpoints(lines)
 
 
 def save_breakpoints(bp_list):
     """
-    :arg bp_list: a list of tuples `(file_name, line)`
+    :arg bp_list: a list of `bdb.Breakpoint` objects
     """
-
     save_path = get_breakpoints_file_name()
     if not save_path:
         return
+
     histfile = open(get_breakpoints_file_name(), 'w')
     bp_list = set([(bp.file, bp.line, bp.cond) for bp in bp_list])
     for bp in bp_list:
